package org.apache.bcel.verifier.structurals;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.Select;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import org.apache.bcel.verifier.structurals.Subroutine;

public class Subroutines {
  private static InstructionHandle[] getSuccessors(InstructionHandle instruction) {
    InstructionHandle[] single = new InstructionHandle[1];
    Instruction inst = instruction.getInstruction();
    if (inst instanceof RET || inst instanceof org.apache.bcel.generic.ReturnInstruction || inst instanceof org.apache.bcel.generic.ATHROW)
      return InstructionHandle.EMPTY_ARRAY; 
    if (inst instanceof JsrInstruction) {
      single[0] = instruction.getNext();
      return single;
    } 
    if (inst instanceof GotoInstruction) {
      single[0] = ((GotoInstruction)inst).getTarget();
      return single;
    } 
    if (inst instanceof BranchInstruction) {
      if (inst instanceof Select) {
        InstructionHandle[] matchTargets = ((Select)inst).getTargets();
        InstructionHandle[] ret = new InstructionHandle[matchTargets.length + 1];
        ret[0] = ((Select)inst).getTarget();
        System.arraycopy(matchTargets, 0, ret, 1, matchTargets.length);
        return ret;
      } 
      InstructionHandle[] pair = new InstructionHandle[2];
      pair[0] = instruction.getNext();
      pair[1] = ((BranchInstruction)inst).getTarget();
      return pair;
    } 
    single[0] = instruction.getNext();
    return single;
  }
  
  private final Map<InstructionHandle, Subroutine> subroutines = new HashMap<>();
  
  public final Subroutine TOPLEVEL;
  
  public Subroutines(MethodGen mg) {
    this(mg, true);
  }
  
  public Subroutines(MethodGen mg, boolean enableJustIceCheck) {
    InstructionHandle[] all = mg.getInstructionList().getInstructionHandles();
    CodeExceptionGen[] handlers = mg.getExceptionHandlers();
    this.TOPLEVEL = new SubroutineImpl(this);
    Set<InstructionHandle> subLeaders = new HashSet<>();
    for (InstructionHandle element : all) {
      Instruction inst = element.getInstruction();
      if (inst instanceof JsrInstruction)
        subLeaders.add(((JsrInstruction)inst).getTarget()); 
    } 
    for (InstructionHandle astore : subLeaders) {
      SubroutineImpl sr = new SubroutineImpl(this);
      sr.setLocalVariable(((ASTORE)astore.getInstruction()).getIndex());
      this.subroutines.put(astore, sr);
    } 
    this.subroutines.put(all[0], this.TOPLEVEL);
    subLeaders.add(all[0]);
    for (InstructionHandle element : all) {
      Instruction inst = element.getInstruction();
      if (inst instanceof JsrInstruction) {
        InstructionHandle leader = ((JsrInstruction)inst).getTarget();
        ((SubroutineImpl)getSubroutine(leader)).addEnteringJsrInstruction(element);
      } 
    } 
    Set<InstructionHandle> instructionsAssigned = new HashSet<>();
    Map<InstructionHandle, ColourConstants> colors = new HashMap<>();
    List<InstructionHandle> qList = new ArrayList<>();
    for (InstructionHandle actual : subLeaders) {
      for (InstructionHandle element : all)
        colors.put(element, ColourConstants.WHITE); 
      colors.put(actual, ColourConstants.GRAY);
      qList.clear();
      qList.add(actual);
      if (actual == all[0])
        for (CodeExceptionGen handler : handlers) {
          colors.put(handler.getHandlerPC(), ColourConstants.GRAY);
          qList.add(handler.getHandlerPC());
        }  
      while (!qList.isEmpty()) {
        InstructionHandle u = qList.remove(0);
        InstructionHandle[] successors = getSuccessors(u);
        for (InstructionHandle successor : successors) {
          if (colors.get(successor) == ColourConstants.WHITE) {
            colors.put(successor, ColourConstants.GRAY);
            qList.add(successor);
          } 
        } 
        colors.put(u, ColourConstants.BLACK);
      } 
      for (InstructionHandle element : all) {
        if (colors.get(element) == ColourConstants.BLACK) {
          ((SubroutineImpl)((actual == all[0]) ? (SubroutineImpl)getTopLevel() : (SubroutineImpl)getSubroutine(actual))).addInstruction(element);
          if (instructionsAssigned.contains(element))
            throw new StructuralCodeConstraintException("Instruction '" + element + "' is part of more than one subroutine (or of the top level and a subroutine)."); 
          instructionsAssigned.add(element);
        } 
      } 
      if (actual != all[0])
        ((SubroutineImpl)getSubroutine(actual)).setLeavingRET(); 
    } 
    if (enableJustIceCheck)
      for (CodeExceptionGen handler : handlers) {
        InstructionHandle _protected = handler.getStartPC();
        while (_protected != handler.getEndPC().getNext()) {
          for (Subroutine sub : this.subroutines.values()) {
            if (sub != this.subroutines.get(all[0]) && sub.contains(_protected))
              throw new StructuralCodeConstraintException("Subroutine instruction '" + _protected + "' is protected by an exception handler, '" + handler + "'. This is forbidden by the JustIce verifier due to its clear definition of subroutines."); 
          } 
          _protected = _protected.getNext();
        } 
      }  
    noRecursiveCalls(getTopLevel(), new HashSet<>());
  }
  
  public Subroutine getSubroutine(InstructionHandle leader) {
    Subroutine ret = this.subroutines.get(leader);
    if (ret == null)
      throw new AssertionViolatedException("Subroutine requested for an InstructionHandle that is not a leader of a subroutine."); 
    if (ret == this.TOPLEVEL)
      throw new AssertionViolatedException("TOPLEVEL special subroutine requested; use getTopLevel()."); 
    return ret;
  }
  
  public Subroutine getTopLevel() {
    return this.TOPLEVEL;
  }
  
  private void noRecursiveCalls(Subroutine sub, Set<Integer> set) {
    Subroutine[] subs = sub.subSubs();
    for (Subroutine sub2 : subs) {
      int index = ((RET)sub2.getLeavingRET().getInstruction()).getIndex();
      if (!set.add(Integer.valueOf(index))) {
        SubroutineImpl si = (SubroutineImpl)sub2;
        throw new StructuralCodeConstraintException("Subroutine with local variable '" + SubroutineImpl.access$000(si) + "', JSRs '" + SubroutineImpl.access$100(si) + "', RET '" + 
            SubroutineImpl.access$200(si) + "' is called by a subroutine which uses the same local variable index as itself; maybe even a recursive call? JustIce's clean definition of a subroutine forbids both.");
      } 
      noRecursiveCalls(sub2, set);
      set.remove(Integer.valueOf(index));
    } 
  }
  
  public Subroutine subroutineOf(InstructionHandle any) {
    for (Subroutine s : this.subroutines.values()) {
      if (s.contains(any))
        return s; 
    } 
    System.err.println("DEBUG: Please verify '" + any.toString(true) + "' lies in dead code.");
    return null;
  }
  
  public String toString() {
    return "---\n" + this.subroutines + "\n---\n";
  }
}
